import json

def to_bytes(x):
    if not isinstance(x, bytes):
        x = x.encode("utf-8")
    return x
    
def to_str(x):
    if isinstance(x, bytes):
        x = x.decode("utf-8")
    return x

class MCError(Exception):
    pass

class WebAPIError(MCError):
    
    Headline = "HTTP error"
    
    def __init__(self, url, response):
        self.URL = url
        self.StatusCode = response.status_code
        self.Message = None
        self.Body = to_str(response.text)
        if response.headers.get("content-type") in ("text/json", "application/json"):
            try:    
                self.Data = json.loads(response.text)
                if isinstance(self.Data, dict):
                    self.Message = self.Data.get("message", "")
            except:
                self.Data = None
        else:
            self.Message = to_str(response.text)
        
    def __str__(self):
        lines = []
        if self.Message:
            lines.append(self.Message)
        else:
            lines.append(self.Body)
        return "\n".join(lines)

    def json(self):
        return self.Data

"""old
class _____ServerSideError(WebAPIError):
    Headline = "Server side application error"

    def __init__(self, url, status_code, type, value):
        message = type
        if value:
            message += ": " + value
        WebAPIError.__init__(self, url, status_code, message=message)
"""

class InvalidArgument(WebAPIError):
    Headline = "Invalid argument"
        
class PermissionError(WebAPIError):
    Headline = "Permission denied"
        
class NotFoundError(WebAPIError):
    Headline = "Object not found"

class BadRequestError(WebAPIError):
    Headline = "Invalid request"




def fetch_generator(c):
    while True:
        tup = c.fetchone()
        if tup is None: break
        yield tup

def chunked(iterable, n):
    if isinstance(iterable, (list, tuple)):
        for i in range(0, len(iterable), n):
            yield iterable[i:i+n]
    else:
        chunk = []
        for item in iterable:
            chunk.append(item)
            if len(chunk) >= n:
                yield chunk
                chunk = []
        if chunk:
            yield chunk
            
def unique(iterable, key=None):
    seen = set()
    for item in iterable:
        k = item if key is None else key(item)
        if not k in seen:
            seen.add(k)
            yield item

def limited(iterable, n):
    if n is None:
        yield from iterable
    if isinstance(iterable, (list, tuple)):
        yield from iterable[:n]
    for f in iterable:
        if n is None:
            yield f
        else:
            if n or n > 0:
                yield f
            else:
                break
            n -= 1
            
def strided(iterable, n, i=0):
    if n is None:
        yield from iterable
    for j, f in enumerate(iterable):
        if j%n == i:
            yield f
            
def skipped(iterable, n):
    if n is None:
        yield from iterable
    if isinstance(iterable, (list, tuple)):
        yield from iterable[n:]
    for f in iterable:
        if n > 0:
            n -= 1
        else:
            yield f



import hashlib
from metacat.util import to_bytes

PasswordHashAlgorithm = "sha1"

def password_hash(user, password):
    hashed = hashlib.new(PasswordHashAlgorithm)
    hashed.update(to_bytes(user))
    hashed.update(b":")
    hashed.update(to_bytes(password))
    return hashed.hexdigest()

def password_digest_hash(realm, user, password):
    # RFC2617 hash of the combination <user, realm, password>
    return hashlib.md5(to_bytes('%s:%s:%s' % (user, realm, password))).digest()



from metacat.util import fetch_generator, chunked
import json, io, csv

def transactioned(method):
    def decorated(first, *params, transaction=None, **args):

        if transaction is not None:
            return method(first, *params, transaction=transaction, **args)
        
        if isinstance(first, HasDB):
            transaction = first.DB.transaction()
        elif isinstance(first, type):
            # class method -- DB is second argument
            transaction = params[0].transaction()
        else:
            transaction = first.transaction()       # static method

        with transaction:
            return method(first, *params, transaction=transaction, **args)

    return decorated

@transactioned
def insert_many(db, table, items, column_names=None, copy_threshold=0, chunk_size=1000, make_tuple=None, transaction=None):
    for chunk in chunked(items, chunk_size):
        if chunk:
            if make_tuple is not None:
                chunk = [make_tuple(item) for item in chunk]
            if len(chunk) <= copy_threshold:
                cols = "" if not column_names else "(" + ",".join(column_names) + ")"
                ncols = len(column_names) if column_names else len(chunk[0])
                vals = ",".join(["%s"] * ncols)
                #print("cols:", cols)
                #print("vals:", vals)
                #print("chunk:", chunk)
                sql = f"insert into {table} {cols} values({vals})"
                #print("sql:", sql)
                transaction.executemany(sql, chunk)
            else:
                csv_file = io.StringIO()
                writer = csv.writer(csv_file, delimiter='\t', quoting=csv.QUOTE_MINIMAL)
                for tup in chunk:
                    assert len(tup) == len(column_names)
                    tup = ["\\N" if x is None else x for x in tup]          # null in Postgres
                    writer.writerow(tup)
                csv_file.seek(0,0)
                transaction.copy_from(csv_file, table, columns = column_names)


class HasDB(object):
    
    def __init__(self, db):
        self.DB = db


class DBObject(HasDB):

    PK = None
    Table = None
    Columns = None

    @classmethod
    def columns(cls, table_name=None, as_text=True, exclude=[]):
        if isinstance(exclude, str):
            exclude = [exclude]
        clist = [c for c in cls.Columns if c not in exclude]
        if table_name:
            clist = [table_name+"."+cn for cn in clist]
        if as_text:
            return ",".join(clist)
        else:
            return clist

    @classmethod
    def list(cls, db):
        c = db.cursor()
        columns = cls.columns()
        c.execute(f"select {columns} from {cls.Table}")
        return (cls.from_tuple(db, tup) for tup in fetch_generator(c))

    @classmethod
    def get(cls, db, *pkvalues):
        assert len(pkvalues) == len(cls.PK)
        wheres = " and ".join([f"{pkc} = %s" for pkc in cls.PK])
        columns = cls.columns()
        sql = f"""
            select {columns}
                from {cls.Table}
                where {wheres}
        """
        c = db.cursor()
        c.execute(sql, pkvalues)
        tup = c.fetchone()
        return None if tup is None else cls.from_tuple(db, tup)
        
    @classmethod
    def exists(cls, db, *pkvalues):
        return cls.get(db, *pkvalues) is not None
        
    def to_json(self):
        return json.dumps(self.to_jsonable())

    @classmethod
    def from_tuple(cls, db, tup):
        return cls(db, *tup)                # default implementstion

    @classmethod
    def from_tuples(cls, db, tuples):
        for tup in tuples:
            yield cls.from_tuple(db, tup)
        
class DBManyToMany(object):
    
    def __init__(self, db, table, *reference_columns, **lookup_values):
        self.DB = db
        self.Table = table
        self.LookupValues = lookup_values
        self.Where = "where " + " and ".join(["%s = '%s'" % (name, value) for name, value in lookup_values.items()])
        assert len(reference_columns) >= 1
        self.ReferenceColumns = list(reference_columns)
        
    def list(self, c=None):
        columns = ",".join(self.ReferenceColumns) 
        if c is None: c = self.DB.cursor()
        c.execute(f"select {columns} from {self.Table} {self.Where}")
        if len(self.ReferenceColumns) == 1:
            return (x for (x,) in fetch_generator(c))
        else:
            return fetch_generator(c)
        
    def __iter__(self):
        return self.list()
        
    def add(self, *vals, c=None):
        assert len(vals) == len(self.ReferenceColumns)
        col_vals = list(zip(self.ReferenceColumns, vals)) + list(self.LookupValues.items())
        cols, vals = zip(*col_vals)
        cols = ",".join(cols)
        vals = ",".join([f"'{v}'" for v in vals])
        if c is None: c = self.DB.cursor()
        c.execute(f"""
            insert into {self.Table}({cols}) values({vals})
                on conflict({cols}) do nothing
        """)
        return self
        
    def contains(self, *vals, c=None):
        assert len(vals) == len(self.ReferenceColumns)
        col_vals = list(zip(self.ReferenceColumns, vals))
        where = self.Where + " and " + " and ".join(["%s='%s'" % (k,v) for k, v in col_vals])
        if c is None: c = self.DB.cursor()
        c.execute(f"""
            select exists(
                    select * from {self.Table} {where} limit 1
            )
        """)
        return c.fetchone()[0]
        
    def __contains__(self, v):
        if not isinstance(v, tuple): v = (v,)
        return self.contains(*v)

    def remove(self, *vals, c=None, all=False):
        assert all or len(vals) == len(self.VarColumns)
        if c is None: c = self.DB.cursor()
        where = self.Where
        if not all:
            col_vals = list(zip(self.ReferenceColumns, vals))
            where += " and " + " and ".join(["%s='%s'" % (k,v) for k, v in col_vals])
        c.execute(f"delete from {self.Table} {where}")
        return self
        
    def set(self, lst, c=None):
        if c is None: c = self.DB.cursor()
        c.execute("begin")
        self.remove(all=True, c=c)
        for tup in lst:
            if not isinstance(tup, tuple):  tup = (tup,)
            self.add(*tup, c=c)
        c.execute("commit")
        



import sys, getopt, os, json, pprint, time, os.path
from textwrap import dedent
from metacat.webapi import MetaCatClient, MCWebAPIError, MCInvalidMetadataError, MCError
from metacat.ui.cli import CLI, CLICommand, InvalidOptions, InvalidArguments
from metacat.util import ObjectSpec, undid
from datetime import timezone, datetime
from .common import load_text, load_file_list, load_json

class DeclareSampleCommand(CLICommand):
    Usage = """-- print sample input for declare-many command
    """
    
    DeclareSample = json.dumps([
        {        
            "namespace":"test",
            "name":"file1.dat",
            "metadata": {
                "pi": 3.14,
                "version":"1.0",
                "format":"raw",
                "done":True
            },
            "size": 1234,
            "parents":[
                {"fid":"4722545"},
                {"did":"my_scope:file_name.data"}, 
                {"namespace":"my_files", "name":"file_name.data"} 
            ]
        },
        {        
            "namespace":"test",
            "name":"file2.dat",
            "metadata": {
                "e": 2.718,
                "version":"1.0",
                "format":"raw",
                "done":False
            },
            "size": 1734
        },
        {        
            "namespace":"test2",
            "auto_name":"files_$clock.dat",
            "metadata": {
                "e": 2.718,
                "version":"1.0",
                "format":"raw",
                "done":False
            },
            "size": 1734
        }
    ], indent=4, sort_keys=True)
    
    def __call__(self, command, client, opts, args):
        print(self.DeclareSample)
        return

class DeclareSingleCommand(CLICommand):
    
    DeclareSample = dedent("""\
        {        
            "metadata": {
                "pi": 3.14,
                "version":"1.0",
                "format":"raw",
                "done":true
            },
            "size": 1234,
            "parents":[
                {"fid":"4722545"},
                {"did":"my_scope:file_name.data"}, 
                {"namespace":"my_files", "name":"file_name.data"} 
            ],
            "checksums": {
                "adler32": "1234abcd"
            }
        }
    """
    )
    
    Opts = ("N:p:m:c:da:s:P:jvf:", ["namespace=", "parents=", "metadata=", "checksums=", "dry-run", "auto-name", "size=", "json",
                    "file-description", "sample", "verbose"])
    Usage = """[options] [[<file namespace>:]<filename>] [<dataset namespace>:]<dataset name>
    Declare signle file:
        declare [options] [[<file namespace>:]<filename>] [<dataset namespace>:]<dataset name>

            -f|--file-description <JSON file>   - JSON file with description, including file attributes and metadata

            The following options can be used to override the values coming from the file description (-f)
            
            -s|--size <size>                    - file size
            -c|--checksums <type>:<value>[,...] - checksums
            -N|--namespace <default namespace>
            -p|--parents <parent>[,...]         - parents can be specified with their file ids or DIDs.
                                                  if the item contains colon ':', it is interpreted as DID
            -m|--metadata <JSON metadata file>  - if unspecified, file will be declared with empty metadata
            -m|--metadata '<JSON metadata>'     - metadata can be specified inline
            -a|--auto-name [[<namespace>:]<pattern>]   - generate file name automatically

            -d|--dry-run                        - dry run: run all the checks but stop short of actual file declaration
            -j|--json                           - print results as JSON
            -v|--verbose                        - verbose output

            --sample                            - print JSON file description sample
            
        Note that file attributes from command line override those from JSON file description. 
        If explicit file name is specified, auto-name is ignored.
    """

    def __call__(self, command, client, opts, args):
        
        if "--sample" in opts:
            print(self.DeclareSample)
            return

        if len(args) == 1:
            file_spec, dataset_spec = None, args[0]
        elif len(args) == 2:
            file_spec, dataset_spec = args
        else:
            raise InvalidArguments("Invalid number of arguments")

        auto_name = file_namespace = file_name = None
        default_namespace = opts.get("-N") or opts.get("--namespace")
        
        dry_run = "-d" in opts or "--dry-run" in opts
        
        file_description = opts.get("-f") or opts.get("--file-description")
        if file_description is not None:
            file_description = json.load(open(file_description, "r"))
        else:
            file_description = {}
        
        size = int(opts.get("-s", opts.get("--size", 0))) or file_description.get("size", 0)
        if size < 0:
            raise InvalidArguments("File size must be non-negative integer")
        file_description["size"] = size

        if file_spec:
            file_namespace, file_name = undid(file_spec, default_namespace)
            file_description["name"] = file_name

        metadata_file = opts.get("-m") or opts.get("--metadata")
        if metadata_file:
            metadata = load_json(metadata_file)
            if not isinstance(metadata, dict):
                raise InvalidArguments("Metadata must be a dictionary")
            file_description["metadata"] = metadata

        dataset_namespace, dataset_name = undid(dataset_spec, default_namespace)
        if not dataset_namespace:
            raise InvalidArguments("Dataset namespace not specified")

        parent_specs = opts.get("-p") or opts.get("--parents")
        if parent_specs:
            parents = []
            for item in parent_specs.split(","):
                if ':' in item:
                    ns, n = undid(item)
                    parents.append({"namespace": ns, "name": n})
                else:
                    parents.append({"fid": item})
            file_description["parents"] = parents

        checksums = opts.get("-c") or opts.get("--checksums")
        if checksums:
            ckdict = {}
            for item in checksums.split(","):
                name, value = item.split(":", 1)
                ckdict[name] = value
            file_description["checksums"] = ckdict

        auto_name_spec = opts.get("-a") or opts.get("--auto-name")
        if auto_name_spec:
            ns, auto_name = undid(auto_name_spec, default_namespace)
            file_namespace = file_namespace or ns
            file_description.pop("name", None)
            file_description["auto_name"] = auto_name

        if file_namespace:
            file_description["namespace"] = file_namespace
            
        if not file_description.get("namespace"):
            file_description["namespace"] = default_namespace

        if not file_description.get("namespace"):
            raise InvalidArguments("File namespace not specified")

        if "-v" in opts or "--verbose" in opts:
            if dry_run:
                print("--- dry run mode ---")
            print(f"File description to be declared and added to dataset {dataset_namespace}:{dataset_name}")
            print(json.dumps(file_description, indent=4, sort_keys=True))
            
        response = list(client.declare_files(f"{dataset_namespace}:{dataset_name}", 
                [file_description], 
                dry_run = dry_run))[0]

        if "-j" in opts or "--json" in opts:
            print(json.dumps(response, indent=4, sort_keys=True))
        else:
            print(response["fid"], response["namespace"], response["name"])

class DeclareManyCommand(CLICommand):
    
    MinArgs = 2
    Opts = ("N:dj", ["namespace=", "dry-run", "json"])
    Usage = """[options] <JSON file with file list> <dataset namespace>:<dataset name>
    Declare multiple files:
            -d|--dry-run                        - dry run: run all the checks but stop short of actual file declaration
            -j|--json                           - print results as JSON
    """

    def __call__(self, command, client, opts, args):
        json_file, dataset_spec = args

        files = json.load(open(json_file, "r"))       # parse to validate JSON

        if ':' not in dataset_spec:
            raise InvalidArguments("Invalid dataset specification")
            
        dataset_namespace, dataset_name = undid(dataset_spec)

        try:
            response = client.declare_files(f"{dataset_namespace}:{dataset_name}", files, dry_run = "-d" in opts)
        except MCError as e:
            print(e)
            sys.exit(1)

        if "-j" in opts or "--json" in opts:
            print(json.dumps(response, indent=4, sort_keys=True))
        else:
            for f in response:
                print(f["fid"], f["namespace"]+':'+f["name"])


class DatasetsCommand(CLICommand):

    Opts = "jp"
    Usage = """[-j|-p] (<file id>|<did>)                   - print datasets containing a file
    
            -p      - pretty-print the list of datasets
            -j      - print the list as JSON
            otherwise print dataset DIDs
    """
    MinArgs = 1

    def __call__(self, command, client, opts, args):
        did = fid = None

        if ':' in args[0]:
            did = args[0]
        else:
            fid = args[0]

        try:
            data = client.get_file(did=did, fid=fid, with_provenance=False, with_metadata=False, with_datasets=True)
        except MCError as e:
            print(e)
            sys.exit(1)

        if data is None:
            print("File not found", file=sys.stderr)
            sys.exit(1)
        datasets = sorted(data.get("datasets", []), key = lambda ds: (ds["namespace"], ds["name"]))
        if "-j" in opts:
            print(json.dumps(datasets, indent=4, sort_keys=True))
        elif "-p" in opts:
            pprint.pprint(datasets)
        else:
            for item in datasets:
                print(item["namespace"] + ":" + item["name"])
                
class FileIDCommand(CLICommand):

    MinArgs = 1
    Usage = """(<namespace>:<name>|<namespace> <name>)  - print file id
    """

    def __call__(self, command, client, opts, args):
        did = namespace = name = None
        if len(args) == 1:
            did = args[0]
            if ':' not in did:
                raise InvalidArguments("Invalid DID: " + did)
        elif len(args) == 2:
            namespace, name = args
        else:
            raise InvalidArguments("Too many arguments")
        
        try:
            data = client.get_file(did=did, namespace=namespace, name=name, with_provenance=False, with_metadata=False,
                               with_datasets=False)
        except MCError as e:
            print(e)
            sys.exit(1)
        if data is None:
            print("File not found", file=sys.stderr)
            sys.exit(1)

        print(data["fid"])

class RetireCommand(CLICommand):

    MinArgs = 1
    Opts = "u unretire"
    Usage = """[-u|--unretire] (<namespace>:<name>|<namespace> <name>)  - retire/unretire file
        -u - unretire the file
    """

    def __call__(self, command, client, opts, args):
        did = namespace = name = None
        if len(args) == 1:
            did = args[0]
            if ':' not in did:
                raise InvalidArguments("Invalid DID: " + did)
        elif len(args) == 2:
            namespace, name = args
        else:
            raise InvalidArguments("Too many arguments")

        try:
            data = client.retire_file(did=did, namespace=namespace, name=name, retire="-u" not in opts)
        except MCError as e:
            print(e)
            sys.exit(1)
        if data is None:
            print("File not found", file=sys.stderr)
            sys.exit(1)

class NameCommand(CLICommand):

    MinArgs = 1
    Opts = "jd json did"
    Usage = """[options] <file id>  - print namespace, name
        -j|--json                   - as JSON {"namespace":..., "name":...}
        -d|--did                    - as DID (namespace:name)
    """

    
    def __call__(self, command, client, opts, args):
        fid = args[0]
    
        try:
            data = client.get_file(fid=fid, with_provenance=False, with_metadata=False, with_datasets=False)
        except MCError as e:
            print(e)
            sys.exit(1)

        if data is None:
            print("File not found", file=sys.stderr)
            sys.exit(1)

        namespace, name = data["namespace"], data["name"]
        if "-j" in opts or "--json" in opts:
            print(f'{{ "namespace": "{namespace}", "name": "{name}" }}')
        elif "-d" in opts or "--did" in opts:
            print(f'{namespace}:{name}')
        else:
            print("Namespace: ", namespace)
            print("Name:      ", name)
        

class ShowCommand(CLICommand):

    Opts = ("mdjpli:", ["json","metadata","pretty","lineage","provenance","id="])
    Usage = """[options] (-i|--id <file id>|<namespace>:<name>)
            -m|--metadata             - include file metadata
            -d|--datasets             - include datasets the file is in

            -j|--json                 - as JSON
            -p|--pretty               - pretty-print information
            
            -l|--provenance           - include provenance information
    """

    
    def __call__(self, command, client, opts, args):
        if not args and "-i" not in opts and "--id" not in opts:
            raise InvalidArguments("Either <namespace:name> or -i|--id <file id> must be specified")

        as_json = "--json" in opts or "-j" in opts
        pretty = "-p" in opts or "--pretty" in opts

        include_provenance = "-l" in opts or "--provenance" in opts
        include_meta = "--metadata" in opts or "-m" in opts
        include_datasets = "--datasets" in opts or "-d" in opts

        did = fid = None
    
        if args:
            did = args[0]
        else:
            fid = opts.get("-i") or opts.get("--id")
            
        if not did and not fid:
            raise InvalidArguments("Eirher DID or file id must be specified")

        try:
            data = client.get_file(did=did, fid=fid, 
                        with_provenance=include_provenance, with_metadata=include_meta,
                        with_datasets=include_datasets)
        except MCError as e:
            print(e)
            sys.exit(1)

        if data is None:
            print("file not found", file=sys.stderr)
            sys.exit(1)
        
        if include_provenance:
            parents = data.get("parents", [])
            if parents:
                parents = client.get_files(parents)
                parents = data["parents"] = [{k:f[k] for k in ("fid", "namespace", "name")} for f in parents]
        
            children = data.get("children", [])
            if children:
                children = client.get_files(children)
                children = data["children"] = [{k:f[k] for k in ("fid", "namespace", "name")} for f in children]

        if as_json:
            print(json.dumps(data, indent=4, sort_keys=True))
        elif pretty:
            pprint.pprint(data)
        else:
            for k, v in sorted(data.items()):
                if k == "checksums":
                    print("checksums:")
                    for typ, cksum in sorted(v.items()):
                        print("    %-10s: %s" % (typ, cksum))
                elif k == "created_timestamp":
                    t = datetime.fromtimestamp(v, timezone.utc)
                    print("%-20s:\t%s" % (k, t))
                elif k not in ("metadata", "parents", "children", "datasets"):
                    print("%-20s:\t%s" % (k, v))

            if "metadata" in data:
                print("metadata:")
                for name, value in sorted(data["metadata"].items()):
                    print("    %-20s: %s" % (name, value))
            if include_provenance:
                if parents:
                    print("parents:")
                    for f in parents:
                        print("   %(namespace)s:%(name)s (%(fid)s)" % f)
                if children:
                    print("children:")
                    for f in children:
                        print("   %(namespace)s:%(name)s (%(fid)s)" % f)
            if "datasets" in data:
                print("datasets:")
                for item in sorted(data["datasets"], key=lambda ds: (ds["namespace"], ds["name"])):
                    print("    %(namespace)s:%(name)s" % item)
                    
class UpdateMetaCommand(CLICommand):
    MinArgs = 1
    Opts = ("i:n:N:rsf:", ["namespace=", "names=", "ids=", "sample", "replace","sample", "files="])
    Usage = """[options] (<JSON file with metadata>|'<JSON expression>')

            -r|--replace          - replace metadata, otherwise update

            list files by DIDs or namespace/names
            -N|--namespace <default namespace>              - default namespace for files
            -f|--files <file namespace>:<file name>[,...]
            -f|--files <file id>[,...]
            -f|--files <file>                               - read the list from text file
            -f|--files <JSON file>                          - read the list from JSON file
            -f|--files -                                    - read the list from stdin
            
            In the text file, a file in the list can be specified as:
                * did               (namespace:name)
                * fid
            In JSON:
                * { "namespace":..., "name":... }
                * { "did":... }
                * { "fid":... }
                * "namespace:name"
    """

    UpdateSample = json.dumps(
        [
            {        
                "did":"test:file1.dat"
            },
            "test:file1.dat",
            {        
                "namespace":"test",
                "name":"file2.dat"
            },
            {        
                "fid":"54634"
            }
        ],
        indent=4, sort_keys=True)
        
    
    def __call__(self, command, client, opts, args):
        
        if "--sample" in opts or "-s" in opts:
            print(self.UpdateSample)
            sys.exit(0)
        
        if len(args) != 1:
            raise InvalidArguments("Invalid arguments")

        # backward compatibility
        opts["-f"] = opts.get("-f") or opts.get("--files") or \
            opts.get("-n") or opts.get("--names") or \
            opts.get("-i") or opts.get("--ids")

        mode = "replace" if ("-r" in opts or "--replace" in opts) else "update"
        namespace = opts.get("-N") or opts.get("--namespace")
    
        file_list = load_file_list(opts["-f"])
        meta = load_json(args[0])

        try:
            response = client.update_file_meta(meta, files=file_list, mode=mode, namespace=namespace)
        except MCError as e:
            print(e)
            sys.exit(1)

class UpdateCommand(CLICommand):
    
    MinArgs = 1
    Opts = "jvdrs:f:k:p:c:m: replace sample verbose json dry-run file-description= size= checksums= parents= children= metadata="
    Usage = """[options] (<file namespace>:<file name>|<file id>)

            -d|--dry-run
            -v|--verbose
            -r|--replace                        - Replace metadata, checksums, parents and children
                                                  otherwise update metadata, checksums, add parents and children.
                                                  Applies to -k, -p, -c, -m, -f options
            -j|--json                           - print updated file attributes as JSON. Otherwise - as Python pprint

            -u|--updates <JSON file>            - JSON file with file attributes to be updated as a dictionary.
                                                  The following keys are accepted: 
                                                      size: int, 
                                                      checksums: dict, 
                                                      metadata: dict,
                                                      parents: list of strings,
                                                      children: list of strings

            -s|--size <size>                    - file size
            -k|--checksums <type>:<value>[,...] - checksums
            -m|--metadata <JSON metadata file>  - metadata
            -m|--metadata '<JSON dictionary>'   - inline metadata
            -p|--parents <parent>[,...]         - parents can be specified with their file ids or DIDs.
                                                  if the item contains colon ':', it is interpreted as DID
            -p|--parents -                      - use '-' with -r to remove all parents
            -c|--children <child>[,...]         - children can be specified with their file ids or DIDs.
                                                  if the item contains colon ':', it is interpreted as DID
            -c|--children -                     - use '-' with -r to remove all choldren
            
            If -u is used together with some individual attributes options, the attributes from the -u file will
            be updated with those coming from the individual attribute options first.
    """

    def __call__(self, command, client, opts, args):
        
        if "--sample" in opts:
            print(self.SampleFile)
            return

        dry_run = "-d" in opts or "--dry-run" in opts
        replce = "-r" in opts or "--replace" in opts
        verbose = "-v" in opts or "--verbose" in opts

        update_args = {}
        updates_file = opts.get("-u", opts.get("--updates"))
        if updates_file:
            update_args = json.load(open(updates_file, "r"))

        update_args["replace"] = "-r" in opts or "--replace" in opts

        if ':' in args[0]:
            update_args["did"] = args[0]
        else:
            update_args["fid"] = args[0]

        size = opts.get("-s", opts.get("--size"))
        if size is not None:
            size = int(size)
            if size < 0:
                raise InvalidArguments("File size must be non-negative integer")
            update_args["size"] = size


        metadata_file = opts.get("-m") or opts.get("--metadata")
        if metadata_file:
            metadata_update = load_json(metadata_file)
            if not isinstance(metadata_update, dict):
                raise InvalidArguments("Metadata file not found or metadata is not a dictionary")
            invalid_names = [k for k in metadata_update if '.' not in k]
            if invalid_names:
                print("Invalid metadata key(s):", ", ".join(invalid_names), file=sys.stderr)
                sys.exit(1)
            update_args["metadata"] = metadata_update

        parents_specs = opts.get("-p") or opts.get("--parents")
        if parents_specs is not None:
            parents = []
            if parents_specs != "-":
                for p in parents_specs.split(","):
                    if ':' in p:
                        parents.append({"did":p})
                    else:
                        parents.append({"fid":p})
            update_args["parents"] = parents

        children_specs = opts.get("-c") or opts.get("--children")
        if children_specs is not None:
            children = []
            if children_specs != "-":
                for c in children_specs.split(","):
                    if ':' in c:
                        children.append({"did":c})
                    else:
                        children.append({"fid":c})
            update_args["children"] = children

        checksums = opts.get("-k") or opts.get("--checksums")
        if checksums:
            ckdict = {}
            for item in checksums.split(","):
                name, value = item.split(":", 1)
                ckdict[name] = value
            update_args["checksums"] = ckdict

        if verbose or dry_run:
            if dry_run:
                print("-------- dry run mode --------")
            for k, v in sorted(update_args.items()):
                print("%-20s: %s" % (k, str(v)))
        
        if not dry_run:
            results = client.update_file(**update_args)
            if verbose:
                print("\nresults:")
            if "--json" in opts or "-j" in opts:
                print(json.dumps(results, sort_keys=True, indent=4))
            elif verbose:
                pprint.pprint(results)
            

class AddCommand(CLICommand):
    
    Opts = ("i:j:n:N:s", ["namespace=", "json=", "names=", "ids=", "sample"])
    Usage = """[options] <dataset namespace>:<dataset name>

            list files by DIDs or namespace/names
            -f|--files <file namespace>:<file name>[,...]
            -f|--files <file id>[,...]
            -f|--files <file>                               - read the list from text file
            -f|--files <JSON file>                          - read the list from JSON file
            -f|--files -                                    - read the list from stdin
            -s|--sample           - print JOSN file list sample
    """
    
    Usage = 'DEPRECATED. Use "metacat dataset add..." instead'
    
    AddSample = json.dumps(
        [
            {        
                "did":"test:file1.dat"
            },
            {        
                "namespace":"test",
                "name":"file2.dat"
            },
            {        
                "fid":"54634"
            }
        ],
        indent=4, sort_keys=True
    )

    
    def __call__(self, command, client, opts, args):
        print('Use "metacat dataset add..." instead')
        sys.exit(1)

        # backward compatibility
        opts["-f"] = (
            opts.get("-f") or opts.get("--files") or
            opts.get("-n") or opts.get("--names") or
            opts.get("-i") or opts.get("--ids") or
            opts.get("-j") or opts.get("--json")
        )

        if opts.get("-n") or opts.get("--names") or \
                        opts.get("-i") or opts.get("--ids") or \
                        opts.get("-j") or opts.get("--json"):
            print("", file=sys.stderr)
            print("Options -j, --json, -n, --names, -i, --ids are deprecated.\nPlease use -f|--files instead", file=sys.stderr)
            print("", file=sys.stderr)


        if "--sample" in opts or "-s" in opts:
            print(self.AddSample)
            sys.exit(0)

        if len(args) != 1:
            raise InvalidArguments("Invalid arguments")

        file_list = load_file_list(opts["-f"])
        dataset = args[-1]
        out = client.add_files(dataset, file_list)
        
class MoveCommand(CLICommand):
    
    Usage = """[options] - move files to a new namespace
    
            -n <target namespace> -f <file list>        - move listed files
            -n <target namespace> -q <query file>       - move selected files
            -n <target namespace> <MQL query>           - move selected files
            
            File list specification
            -f|--files <file namespace>:<file name>[,...]   - list of DIDs
            -f|--files <file id>[,...]                      - list of file ids
            -f|--files <file>                               - read the list of DIDs or file ids from a text file
            -f|--files <JSON file>                          - read the list from JSON file
            -f|--files -                                    - read the list from stdin

            Select files by running a query
            -q|--query <file>                           - read query from the file
            -q|--query -                                - read query from stdin
    """
    Opts = ("f:n:q:", ["namespace=", "files=", "query="])

    def __call__(self, command, client, opts, args):
        
        namespace = opts.get("-n") or opts.get("--namespace")
        if not namespace:
            raise InvalidArguments("Namespace must be specified")
        query = opts.get("-q") or opts.get("--query")
        if query:
            query = load_text(query)
            
        if query and args:
            raise InvalidArguments("Query may be specified using -q|--query or command arguments, but not both")
        
        if not query:
            query = " ".join(args).strip() or None

        file_list = opts.get("-f") or opts.get("--files")
        if file_list:
            file_list = load_file_list(file_list)
        if (file_list is None) == (query is None):
            raise InvalidArguments("Either query or file list must be specified, but not both")
        client.Timeout = None       # this may take long time, so turn the timeout off
        nmoved, errors, nerrors = client.move_files(namespace, file_list=file_list, query=query)
        if errors:
            for error in errors:
                print(error)
            print()
            if len(errors) < nerrors:
                print("Error list was truncated.")
                print("Number of errors:  ", nerrors)
            else:
                print("Number of errors:  ", len(errors))
        print("Files moved:       ", nmoved)
        if errors:
            sys.exit(1)

FileCLI = CLI(
    "declare",  DeclareSingleCommand(),
    "declare-many",  DeclareManyCommand(),
    "declare-sample",  DeclareSampleCommand(),
    "move",     MoveCommand(),
    "add",      AddCommand(),
    "datasets", DatasetsCommand(),
    "update",   UpdateCommand(),
    "update-meta",   UpdateMetaCommand(),
    "retire",   RetireCommand(),
    "name",     NameCommand(),
    "fid",      FileIDCommand(),
    "show",     ShowCommand()
)


from textwrap import dedent, indent


def insert_sql_one(outer, tag, text):
    lines = dedent(outer).split("\n")
    text = dedent(text)
    out_lines = []
    marker = "$" + tag
    for line_no, line in enumerate(lines):
        line = line.expandtabs(4)
        if marker not in line:
            out_lines.append(line)
            continue
        i = line.find(marker)
        text = indent(text, " "*i)
        out_lines += text.split("\n")
        break
    out_lines += lines[line_no+1:]
    return "\n".join(out_lines)
    
def insert_sql(outer, **tags):
    out = outer
    for tag, value in tags.items():
        out = insert_sql_one(out, tag, value)
    return out

def first_not_empty(lst):
    val = None
    for v in lst:
        val = v
        if v is not None and not (isinstance(v, list) and len(v) == 0):
            return v
    else:
        return val
        



from metacat.common.trees import Ascender, Node
from metacat.db import DBFileSet, alias, DBDataset, DBFile
from metacat.common import FileMetaExpressionDNF
from .meta_evaluator import MetaEvaluator
from metacat.util import limited, insert_sql
from textwrap import dedent, indent

class SQLConverter(Ascender):
    
    def __init__(self, db, debug=False, include_retired=False, summary=None):
        self.DB = db
        self.Debug = debug
        self.IncludeRetired = include_retired
        self.Summary = summary

    def columns(self, t, with_meta=True, with_provenance=True):
        meta = f"{t}.metadata" if with_meta else "null as metadata"
        if with_provenance:
            parents = f"{t}.parents"
            children = f"{t}.children"
        else:
            parents = "null as parents"
            children = "null as children"

        attrs = DBFile.attr_columns(alias=t)

        return f"{t}.id, {t}.namespace, {t}.name, {meta}, {attrs}, {parents}, {children}"

    def debug(self, *params, **args):
        if self.Debug:
            parts = ["SQLConverter:"]+list(params)
            print(*parts, **args)

    def __call__(self, tree):
        #print("hello")
        self.debug("\nSQL converter: input tree:----------\n", tree.pretty(), "\n-------------")
        result = self.walk(tree)
        if self.Summary:
            result = Node("summary", [result], mode=mode)
        #print("debug:", self.Debug)
        self.debug("\nSQL converter: output tree:----------\n", result.pretty(), "\n-------------")
        return result

    #
    # Tree node methods
    #
    def query(self, node, *args, namespace=None, name=None):
        raise RuntimeError(f"Named query {namespace}:{name} not assembled")

    def meta_filter(self, node, query=None, meta_exp=None, with_meta=False, with_provenance=False):
        #print("meta_filter: query=", query.pretty())
        #print("meta_filter: node[query]=", node["query"].pretty())
        if meta_exp is None:
            return query
        if query.T == "sql":
            t = alias("t")
            dnf = FileMetaExpressionDNF(meta_exp)
            where_sql = dnf.sql(t)
            if not where_sql:
                return node
            columns = self.columns(t, with_meta, with_provenance)
            query_sql = query["sql"]
            sql = insert_sql(f"""
                -- meta_filter {t}
                    select {columns} 
                    from (
                        $query_sql
                    ) {t} where {where_sql}
                -- end of meta_filter {t}
            """, query_sql=query_sql)
            return Node("sql", sql=sql)
        else:
            return node

    def ordered(self, node, child):
        if child.T == "sql":
            t = alias("t")
            child_sql = child["sql"]
            sql = insert_sql(f"""\
                -- ordered {t}
                    select {t}.*
                    from (
                        $child_sql
                    ) {t} order by {t}.id
                -- end of ordered {t}
            """, child_sql = child_sql)
            return Node("sql", sql=sql)
        elif child.T == "filter":
            return child.clone(ordered=True)
        else:
            return node

    def basic_file_query(self, node, *args, query=None):
        sql = DBFileSet.sql_for_basic_query(self.DB, query, self.IncludeRetired)
        if sql:
            out = Node("sql", sql=sql)
        else:
            out = Node("empty")            # empty file set
        return out

    def file_list(self, node, specs=None, spec_type=None, with_meta=False, with_provenance=False, limit=None, skip=0):
        return Node("sql", sql=DBFileSet.sql_for_file_list(spec_type, specs, with_meta, with_provenance, limit, skip))

    def union(self, node, *args):
        #print("Evaluator.union: args:", args)
        args = [a for a in args if a.T != "empty"]
        if not args:
            return Node("empty")
        sqls = [n for n in args if n.T == "sql"]
        if not sqls:
            return Node("union", args)
        if len(sqls) >= 2:
            inner_sqls = {f"__{i}": n["sql"] for i, n in enumerate(sqls)}
            parts = [f"(\n    $__{i}\n)" for i, n in enumerate(sqls)]
            template = "\nunion\n".join(parts)
            combined_sql = insert_sql(template, **inner_sqls)
            sqls = [Node("sql", sql=combined_sql)]
        others = [n for n in args if n.T != "sql"]
        if not others:
            return sqls[0]
        return Node("union", sqls + others)

    def join(self, node, *args, **kv):
        #print("Evaluator.union: args:", args)
        if any(n.T == "empty" for n in args):
            return Node("empty")
        sqls = [n for n in args if n.T == "sql"]
        if not sqls:
            return node
        if len(sqls) >= 2:
            inner_sqls = {f"__{i}": n["sql"] for i, n in enumerate(sqls)}
            parts = [f"(\n    $__{i}\n)" for i, n in enumerate(sqls)]
            template = "\nintersect\n".join(parts)
            combined_sql = insert_sql(template, **inner_sqls)
            sqls = [Node("sql", sql=combined_sql)]
        others = [n for n in args if n.T != "sql"]
        if not others:
            return sqls[0]
        return Node("join", sqls + others)

    def minus(self, node, *args, **kv):
        #print("Evaluator.union: args:", args)
        assert len(args) == 2
        left, right = args
        if left.T == "empty":
            return Node("empty")
        elif right.T == "empty":
            return left
        if left.T == "sql" and right.T == "sql":
            s1 = left["sql"]
            s2 = right["sql"]
            sql = insert_sql("""\
                (
                    $s1
                )
                except
                (
                    $s2
                )""", s1=s1, s2=s2)
            self.debug("SQLConverter.minus: sql:---------\n", sql, "\n-----------")
            return Node("sql", sql=sql)
        else:
            return node

    def parents_of(self, node, *args, with_meta=False, with_provenance=False, ordered=False):
        assert len(args) == 1
        arg = args[0]
        if arg.T == "empty":    return arg
        if arg.T == "sql":
            with_meta = node["with_meta"]
            with_provenance = node["with_provenance"]
            arg_sql = arg["sql"]
            p = alias("p")
            c = alias("c")
            pc = alias("pc")
            columns = self.columns(p, with_meta, with_provenance)
            order = f"order by {p}.id" if ordered else ""
            table = "files_with_provenance" if with_provenance else "files"
            new_sql = insert_sql(f"""\
                --  parents of {p}
                    select {columns}
                    from {table} {p}
                        inner join parent_child {pc} on {p}.id = {pc}.parent_id
                        inner join (
                            $arg_sql
                        ) as {c} on {c}.id = {pc}.child_id
                    {order}
                --  end of parents of {p}
            """, arg_sql=arg_sql)
            return Node("sql", sql=new_sql)
        else:
            return node

    def children_of(self, node, *args, with_meta=False, with_provenance=False, ordered=False):
        assert len(args) == 1
        arg = args[0]
        if arg.T == "empty":    return arg
        if arg.T == "sql":
            with_meta = node["with_meta"]
            with_provenance = node["with_provenance"]
            arg_sql = arg["sql"]
            p = alias("p")
            c = alias("c")
            pc = alias("pc")
            columns = self.columns(c, with_meta, with_provenance)
            order = f"order by {p}.id" if ordered else ""
            table = "files_with_provenance" if with_provenance else "files"
            new_sql = insert_sql(f"""\
                -- children of {c}
                    select {columns}
                    from {table} {c}
                        inner join parent_child {pc} on {c}.id = {pc}.child_id
                        inner join (
                            $arg_sql
                        ) as {p} on {p}.id = {pc}.parent_id
                    {order}
                -- end of children of {c}
            """, arg_sql=arg_sql)
            return Node("sql", sql=new_sql)
        else:
            return node
            
    def skip_limit(self, node, arg, skip=0, limit=None, **kv):
        if limit is None and skip == 0:
            return arg
        if arg.T == "empty":
            return arg
        if arg.T == "skip_limit":
            new_skip = arg["skip"]
            new_limit = arg["limit"]
            if skip:
                new_skip += skip
                if new_limit is not None:
                    new_limit -= skip
            if limit is not None:
                if new_limit is None:
                    new_limit = limit
                else:
                    new_limit = min(limit, new_limit)
            if new_limit is not None and new_limit <= 0:
                return Node("empty")
            return Node("skip_limit", arg.C, skip=new_skip, limit=new_limit)
        elif arg.T == "sql":
            sql = arg["sql"]
            tmp = alias()
            columns = self.columns(tmp)
            
            limit_clouse = "" if limit is None else f"limit {limit}"
            offset_clouse = "" if skip == 0 else f"offset {skip}"
            
            sql = indent("\n" + sql + "\n", "    ")
            new_sql = insert_sql(f"""\
                -- skip {skip} limit {limit} {tmp}
                    select {columns} 
                    from (
                        $sql
                    ) {tmp} 
                    {limit_clouse} {offset_clouse}
                -- end of limit {limit} {tmp}
            """, sql=sql)
            return Node("sql", sql=new_sql)
        else:
            return node

    def _default(self, node, *children, **named):
        #print("_default:", node.pretty())
        #print("      children:")
        #for c in children:
        #    print("+ ", c.pretty())
        out = Node(node.T, children, _meta=node.M, _data=named)
        #print("_default returning:", out.pretty())
        return out



import time, gzip, os.path
import os, sys
import datetime
from pythreader import PyThread, synchronized, Primitive, TaskQueue, Task
from threading import Timer, Thread

def make_timestamp(t=None):
    if t is None:   
        t = datetime.datetime.now()
    elif isinstance(t, (int, float)):
        t = datetime.datetime.fromtimestamp(t)
    return t.strftime("%m/%d/%Y %H:%M:%S") + ".%03d" % (t.microsecond//1000)

class LogWriter(Primitive):
    
    def __init__(self, name=None):
        Primitive.__init__(self, name=name)

class LogStream(LogWriter):

    def __init__(self, stream, **ignore):
        LogWriter.__init__(self, name=f"LogStream({stream})")
        self.Stream = stream            # sys.stdout, sys.stderr

    @synchronized
    def log(self, msg, raw=False, t=None):
        if t != False and not raw:
            msg = "%s: %s" % (make_timestamp(t), msg)
        self.write(msg + '\n');

    @synchronized
    def write(self, msg):
        self.Stream.write(msg);
        self.Stream.flush()

class CompressTask(Task):
    
    def __init__(self, source):
        Task.__init__(self, name=f"Compress({source})")
        self.Source = source

    def run(self):
        if os.path.isfile(self.Source):
            with open(self.Source, "rb") as inp:
                with gzip.open(self.Source+".gz", "wb") as out:
                    buf = inp.read(10000)
                    while buf:
                        out.write(buf)
                        buf = inp.read(10000)
            os.remove(self.Source)

_CompressQueue = TaskQueue(5)

class LogFile(LogWriter):
        def __init__(self, path, interval = '1d', keep = 10, compress_from = 1, add_timestamp=True, 
                        append=True, flush_interval=None, name=None):
            # interval = 'midnight' means roll over at midnight
            LogWriter.__init__(self, name=f"LogFile({path})")
            self.File = None
            assert isinstance(path, str), "LogFile.__init__: path must be a string. Got %s %s instead" % (type(path), path) 
            self.Path = path
            self.CurLogBegin = 0
            if type(interval) == type(''):
                    mult = 1
                    if interval[-1] == 'd' or interval[-1] == 'D':
                            interval = interval[:-1]
                            mult = 24 * 3600
                            interval = int(interval) * mult
                    elif interval[-1] == 'h' or interval[-1] == 'H':
                            interval = interval[:-1]
                            mult = 3600
                            interval = int(interval) * mult
                    elif interval[-1] == 'm' or interval[-1] == 'M':
                            interval = interval[:-1]
                            mult = 60
                            interval = int(interval) * mult
            self.Interval = interval
            self.Keep = keep
            self.AddTimestamps = add_timestamp
            self.LineBuf = ''
            self.LastLog = None
            self.LastFlush = time.time()
            self.CompressFrom = compress_from
            append = append and os.path.isfile(self.Path)
            if append:
                self.File = open(self.Path, 'a')
                self.File.write("%s: --- log reopened ---\n" % (make_timestamp(),))
                self.CurLogBegin = time.time()
            else:
                self.newLog()
            #print("LogFile: created with file:", self.File)
            if flush_interval is not None:
                self.arm_flush_timer(flush_interval)
                
        def newLog(self):
            if self.File != None:
                self.File.close()
            try:    
                os.remove('%s.%d' % (self.Path, self.Keep))
            except: 
                pass
            try:    
                os.remove('%s.%d.gz' % (self.Path, self.Keep))
            except: 
                pass
            for i in range(self.Keep):
                inx = self.Keep - 1 - i
                old = '%s.%d' % (self.Path, inx) if inx > 0 else self.Path
                new = '%s.%d' % (self.Path, inx + 1)
                try:
                    os.rename(old, new)
                except Exception as e:
                    pass
                try:
                    os.rename(old+".gz", new+".gz")
                except Exception as e:
                    pass
            self.File = open(self.Path, 'w')
            self.CurLogBegin = time.time()
            if self.CompressFrom is not None:
                to_compress = '%s.%d' % (self.Path, self.CompressFrom)
                _CompressQueue << CompressTask(to_compress)

        @synchronized
        def log(self, msg, raw=False, t=None):
            if t is None:   t = time.time()
            if self.Interval == 'midnight':
                if datetime.date.today() != self.LastLog:
                        self.newLog()
            elif isinstance(self.Interval, (int, float)):
                if t > self.CurLogBegin + self.Interval:
                    self.newLog()
            if t != False and not raw:
                msg = "%s: %s" % (make_timestamp(t), msg)
            self._write(msg if raw else msg + "\n")

        @synchronized
        def write(self, msg):
            self.log(msg, raw=True)
            
        @synchronized
        def _write(self, msg):
            if msg:
                #print("LogFile.write: writing to:", self.File)
                self.File.write(msg)
            self.flush()
            self.LastLog = datetime.date.today()

        def arm_flush_timer(self, interval):
            if interval:
                Timer(interval, self.flush).start()
                    
        @synchronized
        def flush(self, interval=None):
            if self.File is not None:
                self.File.flush()
            if interval:
                self.arm_flush_timer(interval)
                
        def start(self):
            # for compatibility with clients, which think LogFile is a thread
            if isinstance(self, PyThread):
                PyThread.start(self)
            elif isinstance(self, Thread):
                Thread.start(self)
            else:
                pass
                
        def __del__(self):
            if self.File is not None:
                self.File.close()
                self.File = None

_LogWriters = {}

def log_writer(output, **args):
    global _LogWriters
    if isinstance(output, LogWriter) or output is None:
        return output
    if output is sys.stdout:
        output = "-"
    elif output is sys.stderr:
        output = "2>"
    assert isinstance(output, str)
    if output not in _LogWriters:
        if output in ("-", "1>"):
            _LogWriters["-"] = LogStream(sys.stdout)
        elif output == "2>":
            _LogWriters["2>"] = LogStream(sys.stderr)
        else:
            _LogWriters[output] = LogFile(output, **args)
    writer = _LogWriters[output]
    return writer
                



import requests, json, fnmatch, sys, os, random, time
from metacat.util import to_str, to_bytes, ObjectSpec, chunked
from metacat.common import SignedToken, TokenLib, TokenAuthClientMixin, AuthenticationError
from urllib.parse import quote_plus, unquote_plus

INVALID_METADATA_ERROR_CODE = 488

def parse_name(name, default_namespace=None):
    words = name.split(":", 1)
    if len(words) < 2:
        assert not not default_namespace, "Null default namespace"
        ns = default_namespace
        name = words[-1]
    else:
        ns, name = words
    return ns, name

undid = parse_name

class MCError(Exception):
    pass

class WebAPIError(MCError):
    
    Headline = "MetaCat API error"
    
    def __init__(self, url, response=None, message=None):
        self.URL = url
        self.StatusCode = response.status_code
        self.Data = None
        self.Body = None
        if message:
            self.Message = self.Body = message
        else:
            self.Body = to_str(response.text)
            if response.headers.get("content-type") in ("text/json", "application/json"):
                try:    
                    self.Data = json.loads(response.text)
                    if isinstance(self.Data, dict):
                        self.Message = self.Data.get("message", "")
                except:
                    self.Data = None
            else:
                self.Message = to_str(response.text)

    def __str__(self):
        lines = []
        if self.Message:
            lines.append(self.Message)
        else:
            lines.append(self.Body)
        return "\n".join(lines)

    def json(self):
        return self.Data
        
class ServerReportedError(WebAPIError):
    Headline = "Server side application error"

    def __init__(self, url, status_code, type, value):
        message = "Type:" + type + f"   Status code:{status_code}"
        if value:
            message += ": " + value
        WebAPIError.__init__(self, url, message=message)
        
class InvalidArgument(WebAPIError):
    Headline = "Invalid argument"
        
class NotFoundError(WebAPIError):
    Headline = "Object not found"

class BadRequestError(WebAPIError):
    Headline = "Invalid request"
            
class InvalidMetadataError(WebAPIError):
    Headline = "Invalid metadata"

    def __str__(self):
        msg = ["Invalid metadata"]
        for item in self.json().get("metadata_errors", []):
            item_headline = item["message"]
            index = item.get("index")
            fid = item.get("fid")
            item_id = ""
            if fid is not None:
                item_id = f"file fid={fid}" + item_id
            if index is not None:
                item_id = f"file #{index} " + item_id
            item_id = item_id.strip()
            item_id = f"{item_id}: " if item_id else ""
            msg.append("  " + item_id + item_headline)
            for error in item.get("metadata_errors", []):
                msg.append("    %s: %s" % (error["name"], error["reason"]))
        return "\n".join(msg)

class HTTPClient(object):

    InitialRetry = 1.0
    RetryExponent = 1.5
    DefaultTimeout = 1800.0

    def __init__(self, server_url, token, timeout=None):
        self.ServerURL = server_url
        self.Token = token
        if timeout is not None and timeout <= 0:
            self.Timeout = None         # no timeout
        else:
            self.Timeout = timeout or self.DefaultTimeout
        self.LastResponse = self.LastURL = self.LastStatusCode = None

    def retry_request(self, method, url, timeout=None, **args):
        """
        Implements the functionality to retry on 503 response with random exponentially growing delay
        Use timemout = 0 to try the request exactly once
        Returns the response with status=503 on timeout
        """
        if timeout is None:
            timeout = self.DefaultTimeout
        tend = time.time() + timeout
        retry_interval = self.InitialRetry
        response = None
        done = False
        while not done:
            if method == "get":
                response = requests.get(url, timeout=self.Timeout, **args)
            else:
                response = requests.post(url, timeout=self.Timeout, **args)
            if response.status_code != 503:
                break
            sleep_time = min(random.random() * retry_interval, tend-time.time())
            retry_interval *= self.RetryExponent
            if sleep_time >= 0:
                time.sleep(sleep_time)
            else:
                break       # time out
        return response

    def send_request(self, method, uri_suffix, headers=None, timeout=None, **args):
        self.LastURL = url = "%s/%s" % (self.ServerURL, uri_suffix)
        default_headers = {
            "Accept": "text/plain, application/json, text/json, application/json-seq"
        }
        if self.Token is not None:
            default_headers["X-Authentication-Token"] = self.Token.encode()
        if headers:
            default_headers.update(headers)
        headers = default_headers
        self.LastResponse = response = self.retry_request(method, url, headers=headers, **args)
        self.LastStatusCode = response.status_code
        if response.status_code == INVALID_METADATA_ERROR_CODE:
            raise InvalidMetadataError(url, response)
        if response.status_code == 404:
            raise NotFoundError(url, response)
        if response.status_code == 400:
            raise BadRequestError(url, response)
        elif response.status_code/100 != 2:
            raise WebAPIError(url, response)
        return response

    def get_text(self, uri_suffix):
        return self.send_request("get", uri_suffix).text

    def post_text(self, uri_suffix, data):
        return self.send_request("post", uri_suffix, data=data).text

    def unpack_json(self, json_text):
        results = json.loads(json_text)
        if isinstance(results, dict):
            if "results" in results:
                results = results["results"]
            elif "error" in results:
                raise ServerReportedError(self.LastURL, self.LastStatusCode, results["error"]["type"], results["error"].get("value", ""))
        return results

    RS = b'\x1E'
    def unpack_json_seq(self, response):
        for line in response.iter_lines():
            if line:    line = line.strip()
            while line and line.startswith(self.RS):
                line = line[1:]
            if line:
                #print(f"stream line:[{line}]")
                obj = self.unpack_json(line)
                yield obj

    def unpack_json_data(self, response):
        response_content_type = response.headers.get("content-type")
        if "application/json-seq" in response_content_type:
            return self.unpack_json_seq(response)
        else:
            return self.unpack_json(response.text)

    def get_json(self, uri_suffix):
        headers = {"Accept": "application/json-seq, application/json, text/json"}
        return self.unpack_json_data(self.send_request("get", uri_suffix, headers=headers, stream=True))

    def post_json(self, uri_suffix, data):
        if not isinstance(data, (str, bytes)):
            data = json.dumps(data)
        headers = {
                "Accept": "application/json-seq, application/json, text/json",
                "Content-Type": "text/json"
        }
        response = self.send_request("post", uri_suffix, data=data, headers=headers, stream=True)
        return self.unpack_json_data(response)

    def get_json_stream(self, uri_suffix):
        url = "%s/%s" % (self.ServerURL, uri_suffix)
        headers = {"Accept": "application/json-seq"}
        if self.Token is not None:
            headers["X-Authentication-Token"] = self.Token.encode()

        response = self.retry_request("get", url, headers=headers, stream=True)
        if response.status_code == INVALID_METADATA_ERROR_CODE:
            raise InvalidMetadataError(url, response)
        if response.status_code == 404:
            raise NotFoundError(url, response)
        elif response.status_code != 200:
            raise WebAPIError(url, response)
        
        if response.headers.get("Content-Type") != "application/json-seq":
            raise WebAPIError(url, response)

        for line in response.iter_lines():
            if line:    line = line.strip()
            while line.startswith(b'\x1E'):
                line = line[1:]
            if line:
                #print(f"stream line:[{line}]")
                obj = json.loads(line)
                yield obj

class MetaCatClient(HTTPClient, TokenAuthClientMixin):
    
    Version = "1.0"
    
    def __init__(self, server_url=None, auth_server_url=None, max_concurrent_queries = 5,
                token = None, token_file = None, token_library = None, timeout = None):    

        """Initializes the MetaCatClient object

        Arguments
        ---------
        server_url : str
            The server endpoint URL, defult = from METACAT_SERVER_URL environment variable 
        auth_server_url : str
            The endpoint URL for the Authentication server, default = server_url + "/auth"
        max_concurrent_queries : int, optional
            Controls the concurrency when asynchronous queries are used
        token_file : str
            File path to read the authentication token from
        token : bytes or str or SignedToken
            Use this token for authentication, optional
        timeout : int or float
            Request timeout in seconds. Default: None - use default timeout, which is 300 seconds
        """

        server_url = server_url or os.environ.get("METACAT_SERVER_URL")
        if not server_url:
            raise RuntimeError("MetaCat server URL unspecified")

        auth_server_url = auth_server_url or os.environ.get("METACAT_AUTH_SERVER_URL")

        TokenAuthClientMixin.__init__(self, server_url, auth_server_url, token=token, token_file=token_file, token_library=token_library)
        HTTPClient.__init__(self, server_url, token=self.token(), timeout=timeout)
        self.MaxConcurrent = max_concurrent_queries
        self.AsyncQueue = None
        
    def sanitize(self, *words, **kw):
        for w in words:
            if "'" in words:
                raise InvalidArgument("", "Invalid value: %s" % (w,))
        for name, value in kw.items():
            if "'" in value:
                raise InvalidArgument("", "Invalid value for %s: %s" % (name, value))
            if "'" in name:
                raise InvalidArgument("", "Invalid name for: %s" % (name,))

    @property
    def async_queue(self):
        if self.AsyncQueue is None:
            try:    
                from pythreader import TaskQueue
            except ModuleNotFoundError:
                raise ModuleNotFoundError("pythreader module required for asynchronous queries. Use: pip install 'pythreader>=2.7.0'")
            self.AsyncQueue = TaskQueue(self.MaxConcurrent)
        return self.AsyncQueue

    def resfresh_token(self):
        if self.TokenFile:
             token = open(self.TokenFile, "rb").read()
             self.Token = SignedToken.decode(token)
        return self.Token
        
    def simulate_503(self):
        return self.get_text("data/simulate_503")

    def get_version(self):
        """Returns server version as text
        """
        return self.get_text("data/version")

    def list_datasets(self, namespace_pattern=None, name_pattern=None, with_counts=False):
        """Gets the list of datasets with namespace/name matching the templates. The templates are
        Python ``fnmatch`` module style templates where ``'*'`` matches any substring and ``'?'`` matches a single character.

        Arguments
        ---------
        namespace_pattern : str
        name_pattern : str
        with_file_counts : boolean
            controls whether the results should include file counts or dataset names only

        Yields
        ------
        generator
            yields dictionaries like {"namespace":..., "name":..., "file_count":...}
        """        
        #url = "data/datasets?with_file_counts=%s" % ("yes" if with_file_counts else "no")
        url = "data/datasets?with_counts=no"
        lst = self.get_json(url)
        promises = []        # [(dataset_dict, promise)]
        for item in lst:
            namespace, name = item["namespace"], item["name"]
            if namespace_pattern is not None and not fnmatch.fnmatch(namespace, namespace_pattern):
                continue
            if name_pattern is not None and not fnmatch.fnmatch(name, name_pattern):
                continue
            if not with_counts:
                yield item
            else:
                # fetch counts asynchronously
                did = namespace + ":" + name
                promises.append(self.async_queue.add(self.get_dataset_counts, did, promise_data = item).promise)

        for promise in promises:
            counts = promise.wait()
            #print("promise: counts:", counts)
            item = promise.Data
            item.update(counts)
            yield item

    def get_dataset_counts(self, did=None, namespace=None, name=None):
        """Gets single dataset files, subsets, supersets, etc. counts
        
        Arguments
        ---------
        did : str - "namespace:name"
        namespace : str
        name : str
        
        Returns
        -------
        dict
            dataset counts or None if the dataset was not found
        """        

        did = ObjectSpec(did, namespace=namespace, name=name).did()
        try:
            out = self.get_json(f"data/dataset_counts?dataset={did}")
            #print("get_dataset_counts", did, " out=", out)
            return out
        except NotFoundError:
            print("get_dataset_counts: None")
            return None

    def get_dataset(self, did=None, namespace=None, name=None, exact_file_count=False):
        """Gets single dataset
        
        Arguments
        ---------
        did : str - "namespace:name"
        namespace : str
        name : str
        
        Returns
        -------
        dict
            dataset attributes or None if the dataset was not found
        """        

        spec = ObjectSpec(did, namespace=namespace, name=name).did()

        try:
            url = f"data/dataset?dataset={spec}"
            if exact_file_count:
                url += "&exact_file_count=yes"
            return self.get_json(url)
        except NotFoundError:
            return None

    def get_dataset_files(self, did, namespace=None, name=None, with_metadata=False, include_retired_files=False):
        """Gets single dataset
        
        Arguments
        ---------
        did : str - "namespace:name"
        namespace : str
        name : str
        
        Returns
        -------
        generator
            generates sequence of dictionaries, one dictionary per file
        """        
        
        if namespace is not None:
            did = namespace + ':' + name
        try:
            with_metadata = "yes" if with_metadata else "no"
            include_retired_files = "yes" if include_retired_files else "no"
            url = f"data/dataset_files?dataset={did}&with_metadata={with_metadata}&include_retired_files={include_retired_files}"
            return self.get_json_stream(url)
        except NotFoundError:
            return None
        
        
    def create_dataset(self, did, frozen=False, monotonic=False, metadata=None, metadata_requirements=None, 
            files_query=None, subsets_query=None,
            description=""):

        """Creates new dataset. Requires client authentication.

        Arguments
        ---------
        did : str
            "namespace:name"
        frozen : bool
        monotonic : bool
        metadata : dict
            Dataset metadata
        metadata_requirements : dict
            Metadata requirements for files in the dataset
        file_query : str
            Run MQL file query and add resulting files to the new dataset
        dataset_query : str
            Run MQL dataset query and add resulting datasets to the new dataset as subsets
        description : str

        Returns
        -------
        dict
            created dataset attributes
        """

        namespace, name = did.split(":",1)
        #self.sanitize(namespace=namespace, name=name)
        params = {
            "namespace":    namespace,
            "name":         name,
            "frozen":       frozen,
            "monotonic":    monotonic,
            "metadata":     metadata or {},
            "metadata_requirements":    metadata_requirements or None,
            "description":  description or "",
            "files_query":  files_query or None,
            "subsets_query":  subsets_query or None
        }
        url = f"data/create_dataset"
        return self.post_json(url, params)
        
    def add_child_dataset(self, parent_spec, child_spec):
        """Adds a child dataset to a dataset.
        
        Arguments
        ---------
        parent_spec : str
            Parent namespace, name ("namespace:name")
        child_spec : str
            Child namespace, name ("namespace:name")
        """
        url = f"data/add_child_dataset?parent={parent_spec}&child={child_spec}"
        return self.get_text(url)
        
    def add_files(self, dataset, file_list=None, namespace=None, query=None):
        """Add existing files to an existing dataset. Requires client authentication.
        
        Arguments
        ---------
        dataset : str
            "namespace:name" or "name", if namespace argument is given
        query : str
            MQL query to run and add files matching the query
        file_list : list
            List of dictionaries, one dictionary per file. Each dictionary must contain either a file id
        
            .. code-block:: python
        
                    { "fid": "abcd12345" }

            or namespace/name:
        
            .. code-block:: python

                    { "name": "filename.data", "namespace": "my_namespace" }

            or DID:
        
            .. code-block:: python

                    { "did": "my_namespace:filename.data" }
        
        namespace : str, optional
            Default namespace. If a ``file_list`` item is specified with a name without a namespace, the ``default namespace``
            will be used.

        Returns
        -------
        int
            number of files added to the dataset
        
        Notes
        -----
        Either ``file_list`` or ``query`` must be specified, but not both
        """
            
        default_namespace = namespace
        if ':' not in dataset:
            if default_namespace is None:
                raise ValueError("Namespace not specified for the target dataset")
            dataset = f"{default_namespace}:{dataset}"

        url = f"data/add_files?dataset={dataset}"
        
        if (file_list is None) == (query is None):
            raise ValueError("Either file_list or query must be specified, but not both")
        
        params = {
            "namespace": namespace,
        }
        if file_list is not None:
            lst = []
            for f in file_list:
                spec = ObjectSpec.from_dict(f, default_namespace)
                spec.validate()
                lst.append(spec.as_dict())
            params["file_list"] = lst
        elif query:
            params["query"] = query
        else:
            raise ValueError("Either file_list or query must be specified, but not both")
        
        out = self.post_json(url, params)
        return out["files_added"]

    def remove_files(self, dataset, file_list=None, namespace=None, query=None):
        """Remove files from a dataset. Requires client authentication.
        
        Arguments
        ---------
        dataset : str
            "namespace:name" or "name", if namespace argument is given
        query : str
            MQL query to run and add files matching the query
        file_list : list
            List of dictionaries, one dictionary per file. Each dictionary must contain either a file id
        
            .. code-block:: python
        
                    { "fid": "abcd12345" }

            or namespace/name:
        
            .. code-block:: python

                    { "name": "filename.data", "namespace": "my_namespace" }

            or DID:
        
            .. code-block:: python

                    { "did": "my_namespace:filename.data" }
        
        namespace : str, optional
            Default namespace. If a ``file_list`` item is specified with a name without a namespace, the ``default namespace``
            will be used.

        Returns
        -------
        int
            actual number of files removed from the dataset
        
        Notes
        -----
        Either ``file_list`` or ``query`` must be specified, but not both
        """
            
        default_namespace = namespace
        dataset_spec = ObjectSpec(dataset, namespace=default_namespace)
        
        if (file_list is None) == (query is None):
            raise ValueError("Either file_list or query must be specified, but not both")
        
        params = {
            "dataset_namespace": dataset_spec.Namespace,
            "dataset_name": dataset_spec.Name,
            "namespace": namespace
        }
        if file_list is not None:
            lst = []
            for f in file_list:
                spec = ObjectSpec.from_dict(f, default_namespace)
                spec.validate()
                lst.append(spec.as_dict())
            params["file_list"] = lst
        elif query:
            params["query"] = query
        else:
            raise ValueError("Either file_list or query must be specified, but not both")
        
        out = self.post_json("data/remove_files", params)
        return out["files_removed"]

    def remove_dataset(self, dataset):
        """Remove a dataset. Requires client authentication.
        
        Arguments
        ---------
        dataset : str
            "namespace:name"
        """
        return self.get_text(f"data/remove_dataset/{dataset}")


    def declare_file(self, did=None, namespace=None, name=None, auto_name=None,
                     dataset_did=None, dataset_namespace=None,
                     dataset_name=None, size=0, metadata={}, fid=None, parents=[], checksums={},
                     dry_run=False):
        """Declare new file and add it to the dataset. Requires client authentication.
        
        Arguments
        ---------
        did : str
            file "namespace:name"
        namespace : str
            file namespace
        name : str
            file name
        auto_name : str
            pattern to use for file name auto generation, default None - do not auto-generate file name
        dataset_did : str
            dataset "namespace:name"
        dataset_namespace : str
            dataset namespace
        dataset_name : str
            dataset name
        size : int
            file size in bytes, default 0
        metadata : dict
            file metadata, default empty dictionary
        fid : str
            file id, default None - to be auto-generated
        checksums : dict
            dictionary with checksum values by the checksum type: {"type":"value", ...}
        parents : list of dicts
            each dict represents one parent file. The dict must contain one of the the following
                - "fid" - parent file id
                - "namespace" and "name" - parent file namespace and name
                - "did" - parent file DID ("<namespace>:<name>")
        dry_run : boolean
            If true, run all the necessary checks but stop short of actual file declaraion or adding to a dataset. 
            If not all checks are successful, generate eirher InvalidMetadataError or WebApiError.
            Default: False = do declare

        Returns
        -------
        dict
            dictionary with file name, namespace and file id. Names and file ids will be auto-generated as necessary.

        Notes
        -----
        At least one of the following must be specified for the file:
            - did
            - namespace and either name or auto_name

        At least one of the following must be specified for the dataset:
            - dataset_did
            - dataset_namespace and dataset_name

        Auto-name pattern can be any string with the following substrings, which will be replaced with appropriate values to generate the file name:

            - $clock - current interger timestamp in milliseconds
            - $clock3 - last 3 digits of $clock - milliseconds only
            - $clock6 - last 6 digits of $clock
            - $clock9 - last 9 digits of $clock
            - $uuid - random UUID in hexadecimal representation, 32 hex digits
            - $uuid16 - 16 hex digits from random UUID hexadecimal representation
            - $uuid8 - 8 hex digits from random UUID hexadecimal representation
            - $fid - file id
        """

        if not did:
            if not namespace:
                raise ValueError("Unspecified file namespace")
            if not name and not auto_name:
                raise ValueError("Unspecified file name")
        else:
            namespace, name = undid(did)
        if not (dataset_namespace and dataset_name) and not dataset_did:
            raise ValueError("Either dataset_did or dataset_namespace and dataset_name must be provided")
        if dataset_did is None:
            dataset_did = f"{dataset_namespace}:{dataset_name}" 
        info = dict(
            namespace = namespace, 
            name = name,
            size = size,
            checksums = checksums,
            fid = fid,
            parents = parents,
            metadata = metadata
        )
        if not name and auto_name:
            info["auto_name"] = auto_name
        return self.declare_files(dataset_did, [info])[0]

    def declare_files(self, dataset, files, namespace=None, dry_run=False):
        """Declare new files and add them to an existing dataset. Requires client authentication.
        
        Arguments
        ---------
        dataset : str
            "namespace:name"
        files : list or dict
            List of dictionaries, one dictionary per a file to be declared. See Notes below for the expected contents of each
            dictionary.
            For convenience, if declaring single file, the argument can be the single file dictionary instead of a list.
        namespace: str, optional
            Default namespace for files to be declared
        dry_run : boolean
            If true, run all the necessary checks but stop short of actual file declaraion or adding to a dataset. 
            If not all checks are successful, generate eirher InvalidMetadataError or WebApiError.
            Default: False = do declare

        Returns
        -------
        list
            list of dictionaries, one dictionary per file with file ids: { "fid": "..." }
        
        Notes
        -----
        Each file to be declared must be represented with a dictionary. The dictionary must contain one of:
        
        "did" - string in the format "<namespace>:<name>"

        "name" - file name and optionaly "namespace". If namespace is not present, the ``namespace`` argument will be used
                 as the default namespace

        "auto_name" - pattern to auto-generate file name
    
        .. code-block:: python
    
            { 
                "namespace": "namespace",           # optional, namespace can be specified for each file explicitly or implicitly using the namespace=... argument
                "name": "filename",                 # optional,
                "did": "namespace:filename",        # optional, convenience for Rucio users
                                                    # either "did" or "name", "namespace" must be present
                "size": ...,                        # required, integer number of bytes
                "metadata": {...},                  # optional, file metadata, a dictionary with arbitrary JSON'able contents
                "fid":  "...",                      # optional, file id. Will be auto-generated if unspecified.
                                                    # if specified, must be unique
                "parents": [...],                   # optional, list of dicts, one dict per parent. See below.
                "checksums": {                      # optional, checksums dictionary
                    "method": "value",...
                },
                "auto_name": "..."                  # optional, pattern to auto-generate file name if name is not specified or null
            },...
    
        Parents are specified with dictionaries, one dictionary per file. Each dictionary specifies the parent file in one of three ways:

            - "did": "<namespace>:<name>"
            - "namespace":"...", "name":"..."
            - "fid": "<file id>"
    
        DEPRECATED: if the parent is specified with a string instead of a dictionary, it is interpreferd as the parent file id.
        """        
        
        default_namespace = namespace
        if isinstance(files, dict):
            files = [files]                     # convenience

        lst = []

        for i, item in enumerate(files):
            f = item.copy()
            namespace = f.get("namespace", default_namespace)
            if "did" in f:
                if "name" in f or "namespace" in f:
                    raise ValueError(f"Both DID and namespace/name specified for {did}")
                did = f.pop("did")
                namespace, name = parse_name(did, default_namespace)
                f["name"] = name
            f["namespace"] = namespace
            size = f.get("size")
            if not isinstance(size, int) or size < 0:
                raise ValueError(f"File size is unspecified or invalid for file #{i} in the list")

            meta = item.get("metadata", {})
            for k in meta.keys():
                if '.' not in k:
                    raise ValueError(f'Invalid metadata key "{k}" for file #{i} in the list: metadata key must contain dot (.)')

            f["metadata"] = meta
            lst.append(f)

        url = f"data/declare_files?dataset={dataset}"
        if dry_run: url += "&dry_run=yes"
        out = self.post_json(url, lst)
        return out
        
    def move_files(self, namespace, file_list=None, query=None):
        """
        Arguments
        ---------
        namespace : str
            namespace to move files to
        query : str
            MQL query to run and add files matching the query
        file_list : list
            List of dictionaries, one dictionary per file. Each dictionary must contain either a file id
        
            .. code-block:: python
        
                    { "fid": "abcd12345" }

            or namespace/name:
        
            .. code-block:: python

                    { "name": "filename.data", "namespace": "my_namespace" }

            or DID:
        
            .. code-block:: python

                    { "did": "my_namespace:filename.data" }
        Returns
        -------
        tuple
            number of files moved, list of errors, if any
        """    
        params = {
            "namespace": namespace,
        }
        if file_list is not None:
            lst = []
            for f in file_list:
                spec = ObjectSpec.from_dict(f)
                spec.validate()
                lst.append(spec.as_dict())
            params["files"] = lst
        elif query:
            params["query"] = query
        else:
            raise ValueError("Either file_list or query must be specified, but not both")

        url = "data/move_files"
        out = self.post_json(url, params)
        errors = out.get("errors", [])
        return out["files_moved"], errors, out.get("nerrors", len(errors))

    def update_file(self, did=None, namespace=None, name=None, fid=None, replace=False,
                size=None, checksums=None, parents=None, children=None, metadata=None
        ):
        """
        Arguments
        ---------
        did : str
            file "namespace:name"
        fid : str
            file id
        namespace : str
            file namespace
        name : str
            file name
        replace : bool
            If True, the specified attribute values will be replaced with new values. 
            Otherwise added (for parents and children) and updated (for checksums and metadata)
        size : int >= 0
            file size, optional
        checksums : dict
            checksum values, optional
        parents : list
            list of parent file ids, optional
        children : list
            list of child file ids, optional
        metadata : dict
            dictionary with metadata to update or replace, optional

        Returns
        -------
        dict
            Dictionary with updated file information
        """

        data = {"mode":"replace" if replace else "add-update"}

        if fid:
            data["fid"] = fid
        else:
            if did:
                namespace, name = did.split(':', 1)
            assert namespace and name
            data["namespace"] = namespace
            data["name"] = name
            
        if size is not None:
            assert isinstance(size, int) and size >= 0
            data["size"] = size
            
        if checksums is not None:
            assert isinstance(checksums, dict)
            data["checksums"] = checksums
            
        if parents is not None:
            assert isinstance(parents, list)
            data["parents"] = [ObjectSpec(p).as_dict() for p in parents]
            
        if children is not None:
            assert isinstance(children, list)
            data["children"] = [ObjectSpec(c).as_dict() for c in children]
            
        if metadata is not None:
            assert isinstance(metadata, dict)
            data["metadata"] = metadata

        return self.post_json("data/update_file", data)

    def update_file_meta(self, metadata, files=None, names=None, fids=None, namespace=None, dids=None, mode="update"):
        """Updates metadata for existing files. Requires client authentication.
        
        **DEPRECATED** *update_file() should be used instead*
        
        Arguments
        ---------
        metadata : dict
            see Notes
        files : list of dicts
            Each dict specifies a file. See Notes
        names : list of strings
            List of file names. Requires namespace to be specified
        dids : list of strings
            List of DIDs ("namespace:name") strings
        fids : list of strings
            List of file ids. The list of files can be specified with ``fids`` or with ``names`` argument, but not
            both.
        namespace : string
            Default namespace
        mode : str
            Either ``"update"`` (default) or ``"replace"``. If mode is ``"update"``, existing metadata will be updated with
            values in ``metadata``. If ``"replace"``, then new values will replace existing metadata. Also, see notes below.
        
        Returns
        -------
        list
            list of dictionaries, one dictionary per file with file ids: { "fid": "..." }
        
        
        Notes
        -----
        This method can be be used to apply common metadata changes to a list of files. This method **can not** be used to update
        file provenance information.
        
        The``metadata`` argument is used to specify the common changes to the metadata to apply to multiple files.
        The ``metadata`` dictionary will be used to either update existing metadata of listed files (if ``mode="update"``) or
        replace it (if ``mode="replace"``).
        
        Files to update have to be specified in one of the following ways:
        
            - files = [list of dicts] - each dict must be in one of the following formats:
        
                - {"fid":"<file id>"} 
                - {"namespace":"<file namespace>", "name":"<file name>"} - namespace is optional. Default: the value of the "namespace" method argument
                - {"did":"<file namespace>:<file name>"} 
        
            - dids = [list of file DIDs]
            - names = [list of file names] - "namespace" argument method must be used to specify the common namespace
            - fids = [list of file ids]
        """
        
        if names and not namespace:
            raise ValueError("Namespace must be specified with names argument")

        def combined():
            for name in (names or []):
                yield ObjectSpec(namespace, name).as_dict()
            for did in (dids or []):
                spec = ObjectSpec(did)
                spec.validate()             # will raise ValueError
                yield spec.as_dict()
            for fid in (fids or []):
                yield ObjectSpec(fid=fid).as_dict()
            for item in (files or []):
                spec = ObjectSpec.from_dict(item)
                spec.validate()             # will raise ValueError
                yield spec.as_dict()

        url = f"data/update_file_meta"
        out = []
        for chunk in chunked(combined(), 1000):
            data = {
                "metadata":metadata,
                "files":chunk,
                "mode":mode
            }
            out.extend(self.post_json(url, data))

        return out

    def delete_file(self, did=None, namespace=None, name=None, fid=None):
        """Delete an existing file. The file will be removed from all datasets and the database and its name and file id can be reused.
        
        Arguments
        ---------
        did : str
            file "namespace:name"
        fid : str
            file id
        namespace : str
            file namespace
        name : str
            file name
        retire : bool
            whether the file should be retired
        """
        data = {}
        if fid:
            data["fid"] = fid
        else:
            if did:
                namespace, name = did.split(':', 1)
            assert namespace and name
            data["namespace"] = namespace
            data["name"] = name
        #print("API.delete: sending:", data)
        return self.post_json("data/delete_file", data)

    def retire_file(self, did=None, namespace=None, name=None, fid=None, retire=True):
        """Modify retired status of the file. Retured file remains in the database, "occupies" the name in the namespace, but
        id not visible to normal queries. Retired file can be brought back to normal using this method too.
        
        If you need to completely remove the file, use `delete_file` method.
        
        Arguments
        ---------
        did : str
            file "namespace:name"
        fid : str
            file id
        namespace : str
            file namespace
        name : str
            file name
        retire : bool
            whether the file should be retired
        
        Returns
        -------
        dict
            Dictionary with updated file information
        """
        data = {
            "retire":   retire
        }
        if fid:
            data["fid"] = fid
        else:
            if did:
                namespace, name = did.split(':', 1)
            assert namespace and name
            data["namespace"] = namespace
            data["name"] = name
        #print("API.retire: sending:", data)
        return self.post_json("data/retire_file", data)

    def update_dataset(self, dataset, metadata=None, mode="update", frozen=None, monotonic=None, description=None):   
        """Update dataset. Requires client authentication.
        
        Arguments
        ---------
        dataset : str
           "namespace:name"
        metadata : dict or None
            New metadata values, or, if None, leave the metadata unchanged
        mode: str
            Either ``"update"`` or ``"replace"``. If ``"update"``, metadata will be updated with new values. If ``"replace"``,
            metadata will be replaced with new values.
            If ``metadata`` is None, ``mode`` is ignored
        frozen: boolean or None
            if boolean, new value for the flag. If None, leave it unchanged
        monotonic: boolean or None
            if boolean, new value for the flag. If None, leave it unchanged
        description: str or None
            if str, new dataset description. If None, leave the description unchanged

        Returns
        -------
        dict
            dictionary with new dataset information
        """
        request_data = {}
        if metadata is not None:
            request_data["mode"] = mode
            request_data["metadata"] = metadata
        if frozen is not None:  request_data["frozen"] = frozen
        if monotonic is not None:  request_data["monotonic"] = monotonic
        if description is not None: request_data["description"] = description
        url = f"data/update_dataset?dataset={dataset}"
        out = self.post_json(url, request_data)
        return out
        
    def get_files(self, lookup_list, with_metadata = True, with_provenance=True):
        """Get many file records
        
        Arguments
        ---------
        lookup_list : list
            List of dictionaries, one dictionary per file. Each dictionary must have either
                "did":"namespace:name", or
                "namespace":"..." and "name":"..." or
                "fid":"file id"
        with_metadata : boolean
            whether to include file metadata
        with_provenance:
            whether to include parents and children list

        Returns
        -------
        List of file records, each record is the same as returned by get_file()
        """
        
        with_metadata = "yes" if with_metadata else "no"
        with_provenance = "yes" if with_provenance else "no"
        
        #print("with_metadata:", with_metadata)
        
        new_list = []
        for item in lookup_list:
            if "fid" in item or "namespace" in item and "name" in item:
                pass
            elif "did" in item:
                did = item["did"]
                try:
                    namespace, name = did.split(':', 1)
                except ValueError:
                    raise ValueError("Invalid DID format: " + did)
                item = {"namespace":namespace, "name":name}
            else:
                raise ValueError("Invalid file specifification: " + str(item))
            new_list.append(item)

        url = "data/files?with_metadata=%s&with_provenance=%s" % (with_metadata, with_provenance)

        return self.post_json(url, new_list) 
        
    def get_file(self, name=None, namespace=None, fid=None, did=None, with_metadata = True, with_provenance=True, with_datasets=False):
        """Get one file record
        
        Arguments
        ---------
        fid : str, optional
            File id
        name : str, optional
        namespace : str, optional
            name and namespace must be specified together
        did : str, optional
            "nemaspace:name"
        with_metadata : boolean
            whether to include file metadata
        with_provenance : boolean
            whether to include parents and children list
        with_datasets : boolean
            whether to include the list of datasets the file is in

        Returns
        -------
        dict
            dictionary with file information or None if the file was not found

            .. code-block:: python

                {       
                    "name": "namespace:filename",       # file name, namespace
                    "fid":  "...",                      # files id
                    "creator":  "...",                  # username of the file creator
                    "created_timestamp":   ...,         # numeric UNIX timestamp
                    "size": ...,                        # file size in bytes
                    "checksums": { ... },               # file checksums

                    # included if with_provenance=True
                    "parents":  ["fid",...],            # list of ids for the file parent files
                    "children": ["fid",...],            # list of ids for the file child files

                    # included if with_metadata=True
                    "metadata": { ... },                # file metadata

                    # included if with_datasets=True
                    "datasets": [
                        {"namespace":"...", "name":"..."}, ...
                    ]
                }

        Notes
        -----
        Retrieving file provenance and metadata takes slightly longer time
        """        
        assert (fid is not None) or (did is not None) or (name is not None and namespace is not None), \
            "Either DID or file id or namespace and name must be specified"
        with_meta = "yes" if with_metadata else "no"
        with_rels = "yes" if with_provenance else "no"
        with_datasets = "yes" if with_datasets else "no"
        url = f"data/file?with_metadata={with_meta}&with_provenance={with_rels}&with_datasets={with_datasets}"
        if did:
            namespace, name = parse_name(did, None)
            assert namespace is not None, f"Invalid DID format: {did}"
        if name:
            url += f"&name={name}&namespace={namespace}"
        else:
            url += f"&fid={fid}"        
            
        try: 
            return self.get_json(url)
        except NotFoundError:
            return None

    def query(self, query, namespace=None, with_metadata=False, with_provenance=False, save_as=None, add_to=None,
                        include_retired_files=False, summary=None):
        """Run file query. Requires client authentication if save_as or add_to are used.
        
        Arguments
        ---------
        query : str
            Query in MQL
        namespace : str
            default namespace for the query
        include_retired_files:
            boolean, whether to include retired files into the query results, default=False
        with_metadata : boolean
            whether to return file metadata
        with_provenance : boolean
            whether to return parents and children list
        save_as : str
            namespace:name for a new dataset to create and add found files to
        add_to : str
            namespace:name for an existing dataset to add found files to
        summary : str or None
            "count" - return file count only as int
            "keys" - return list of all top level metadata keys for the selected files
            ``summary`` can not be used together with ``save_as`` or ``add_to``

        Returns
        -------
        list of dicts
            dictionary with file information. Each file will be represented with a dictionary in this list.
        
        Notes
        -----
        Retrieving file provenance and metadata takes slightly longer time
        """
        
        assert not (summary is not None and (add_to or save_as)), "Summary can not be used together with add_to or save_as"
        assert summary in ("count", "keys", None)
        
        if summary:
            url = f"data/query?summary={summary}"
            if namespace:
                url += f"&namespace={namespace}"
            if include_retired_files:
                url += "&include_retired_files=yes"
        else:
            url = "data/query?with_meta=%s&with_provenance=%s" % ("yes" if with_metadata else "no","yes" if with_provenance else "no")
            if namespace:
                url += f"&namespace={namespace}"
            if save_as:
                url += f"&save_as={save_as}"
            if add_to:
                url += f"&add_to={add_to}"
            if include_retired_files:
                url += "&include_retired_files=yes"
        #print("url:", url)
        results = self.post_json(url, query)
        return results

    def async_query(self, query, data=None, **args):
        """Run the query asynchronously. Requires client authentication if save_as or add_to are used.
        
        Arguments
        ---------
        query : str
            Query in MQL
        data : anything
            Arbitrary data associated with this query
        args : 
            Same keyword arguments as for the query() method
        
        Returns
        -------
        Promise
            ``pythreader`` Promise object associated with this query. The promise object will have Data attribute containig the object passed as the ``data``
            argument to the ``async_query`` call. 
        
            See notes below for more on how to use this method.
        """
        
        return self.async_queue.add(self.query, query, promise_data=data, **args).promise

    def wait_queries(self):
        """
        Wait for all issued asynchronous queries to complete
        """
        self.async_queue.waitUntilEmpty()
    
    def create_namespace(self, name, owner_role=None, description=None):
        """Creates new namespace. Requires client authentication.
        
        Arguments
        ---------
        name : str
            Namespace name
        owner_role : str
            Owner role for the new namespace. The user must be a member of the role.
            Optional. If unspecified, the new namespace will be owned by the user.
        description : str
            New namespace description

        Returns
        -------
        dict 
            New namespace information
        """
        
        url = f"data/create_namespace?name={name}"
        if owner_role:
            url += f"&owner_role={owner_role}"
        if description:
            desc = quote_plus(description)
            url += f"&description={desc}"
        return self.get_json(url)
        
    def get_namespace(self, name):
        """Get information about a snamespace
        
        Arguments
        ---------
        name : str
            Namespace name

        Returns
        -------
        dict 
            Namespace information or None if the namespace was not found
        """
        
        try: 
            return self.get_json(f"data/namespace?name={name}")
        except NotFoundError:
            return None
        
    def get_namespaces(self, names):
        """Get information for multiple namespaces
        
        Arguments
        ---------
        names : list of str
            Namespace names

        Returns
        -------
        list 
            Namespace information
        """
        
        return self.post_json(f"data/namespaces", names)
        
    def list_namespaces(self, pattern=None, owner_user=None, owner_role=None, directly=False):
        """List namespaces
        
        Arguments
        ---------
        pattern : str
            Optional fnmatch style pattern to filter namespaces by name
        owner_user : str
            Optional, return only namespaces owned by the specified user
        directly : boolean
            If False and owner_user is specified, return also namespaces owned by all roles the user is in
            Ignored if owner_user is not specified
        owner_role : str
            Optional, return only namespaces owned by the specified role.
            Ignored if owner_user is also specified

        Returns
        -------
        list 
            List of dictionaries with namespace information sorted by the namespace name
        """
        url = "data/namespaces"
        args = ""
        if owner_user:
            args += f"owner_user={owner_user}"
            if directly:
                args += "&directly=yes"
        if owner_role:
            if args: args += "&"            # low level API on the server side will ignore owner_role if owner_user is present, but pass both anyway
            args += f"owner_role={owner_role}"
        if args:
            args = '?' + args
        lst = self.get_json("data/namespaces" + args)
        for item in lst:
            if pattern is None or fnmatch.fnmatch(item["name"], pattern):
                yield item
                
    #
    # Categiries
    #
    def list_categories(self, root=None):
        """List namespaces

        Arguments
        ---------
        root : str
            Optional, if present, list only categories under the root

        Returns
        -------
        list 
            List of dictionaries with category information sorted by category path
        """
        lst = self.get_json("data/categories")
        if root:
            if not root.endswith('.'):
                root += '.'
            lst = [cat for cat in lst if cat["path"].startswith(root)]
        return sorted(lst, key=lambda c: c["path"])

    def get_category(self, path):
        """Get category information
        
        Returns
        -------
        dict 
            A dictionary with category information or None if not found
        """
        out = self.get_json(f"data/category/{path}")
        return out
        
    #
    # Named queries
    #
    def get_named_query(self, namespace, name):
        """Get named query
        
        Arguments
        ---------
        namespace : str
        name : str
        
        Returns
        -------
        dict or None
            A dictionary with information about the named query or None if the named query does not exist.
        """
        try:    data = self.get_json(f"data/named_query?namespace={namespace}&name={name}")
        except NotFoundError:
            return None
        return data

    def list_named_queries(self, namespace=None):
        """Get multiple named queries
        
        Arguments
        ---------
        namespace : str
            optional, if specified the list will include all named queries in the namespace. Orherwise all named queries will be returned
        
        Returns
        -------
        list
            List of dictionaries with information about the named queries.
        """
        url = "data/named_queries"
        if namespace is not None:
            url += f"?namespace={namespace}"
        return self.get_json(url)

    def create_named_query(self, namespace, name, source, parameters=[], update=False):
        data = dict(namespace=namespace, name=name, source=source, parameters=parameters)
        url = "data/create_named_query"
        if update: url += "?update=yes"
        return self.post_json(url, data)



import itertools, io, csv, json
from psycopg2 import IntegrityError

Debug = False

def debug(*parts):
    if Debug:
        print(*parts)
        
Aliases = {}
def alias(prefix="t"):
    global Aliases
    i = Aliases.get(prefix, 1)
    Aliases[prefix] = i+1
    return f"{prefix}_{i}"

class AlreadyExistsError(Exception):
    pass

class DatasetCircularDependencyDetected(Exception):
    pass


class NotFoundError(Exception):
    def __init__(self, msg):
        self.Message = msg

    def __str__(self):
        return "Not found error: %s" % (self.Message,)


def parse_name(name, default_namespace=None):
    words = (name or "").split(":", 1)
    if not words or not words[0]:
        assert not not default_namespace, "Null default namespace"
        ns = default_namespace
        name = words[-1]
    else:
        assert len(words) == 2, "Invalid namespace:name specification:" + name
        ns, name = words
    return ns, name


class MetaValidationError(Exception):
    
    def __init__(self, message, errors):
        self.Errors = errors
        self.Message = message
        
    def as_json(self):
        return json.dumps(
            {
                "message":self.Message,
                "metadata_errors":self.Errors
            }
        )
        
def make_list_if_short(iterable, limit):
    # convert iterable to list if it is short. otherwise return another iterable with the same elements
    
    if isinstance(iterable, (list, tuple)):
        return iterable, None
    
    head = []
    if len(head) < limit:
        for x in iterable:
            head.append(x)
            if len(head) > limit:
                return None, itertools.chain(head, iterable)
        else:
            return head, None
    else:
        return None, iterable



from .http_client import HTTPClient
from .exceptions import MCError, NotFoundError, InvalidArgument, PermissionError, BadRequestError, WebAPIError
from .meta_dnf import FileMetaExpressionDNF, DatasetMetaExpressionDNF

from .rfc2617 import digest_client, digest_server
from .signed_token_jwt import SignedToken, SignedTokenExpiredError, SignedTokenImmatureError, \
        SignedTokenUnacceptedAlgorithmError, SignedTokenSignatureVerificationError
from .token_lib import TokenLib
from .dbbase import DBObject, DBManyToMany, transactioned, insert_many
from .password_hash import password_hash, PasswordHashAlgorithm, password_digest_hash
from .auth_client import TokenAuthClientMixin, AuthenticationError
from .attributes import FileAttributes, DatasetAttributes
from .transaction import ConnectionWithTransactions, Transaction



import json
from .authenticators import authenticator
from metacat.common import DBObject, DBManyToMany, password_digest_hash, transactioned
from metacat.util import fetch_generator

class DBAuthenticator(DBObject):
    
    Table = "authenticators"
    Columns = ["username", "type", "issuer", "user_info"]
    PK = ["username", "type", "issuer"]
    
    def __init__(self, db, username, type, issuer, info):
        DBObject.__init__(self, db)
        self.Username = username
        self.Type = type
        self.Issuer = issuer
        self.Info = info

    @staticmethod
    def list(db, username=None, type=None, issuer=None):
        table = DBAuthenticator.Table
        c = db.cursor()
        columns = DBAuthenticator.columns()
        c.execute(f"""
            select {columns}
                from {table}
                where (%(username)s is null or username=%(username)s)
                    and (%(type)s is null or username=%(type)s)
                    and (%(issuer)s is null or username=%(issuer)s)
        """, dict(username=username, type=type, issuer=issuer))
        return (DBAuthenticator.from_tuple(db, tup) for tup in fetch_generator(c))


class BaseDBUser(DBObject):
    
    Table = "users"
    Columns = "username,name,email,flags,auth_info,auid".split(",")
    PK = ["username"]

    def __init__(self, db, username, name, email, flags, auth_info, auid):
        DBObject.__init__(self, db)
        self.Username = username
        self.Name = name
        self.AUID = auid
        self.EMail = email
        self.Flags = flags
        self.AuthInfo = auth_info or {}        # type -> [secret,...]        # DB representation
        self.RoleNames = None
        
    def __str__(self):
        return "BaseDBUser(%s, %s, %s, %s, %s)" % (self.Username, self.Name, self.EMail, self.Flags, self.AUID)
        
    __repr__ = __str__
    
    @transactioned
    def save(self, transaction=None):
        auth_info = json.dumps(self.AuthInfo)
        columns = self.columns()
        transaction.execute(f"""
            insert into users({columns}) values(%s, %s, %s, %s, %s, %s)
                on conflict(username) 
                    do update set name=%s, email=%s, flags=%s, auth_info=%s, auid=%s;
            """,
            (self.Username, self.Name, self.EMail, self.Flags, auth_info, self.AUID,
                            self.Name, self.EMail, self.Flags, auth_info, self.AUID
            ))
        return self

    def authenticate(self, method, auth_config, presented):
        a = authenticator(method, auth_config, self.AuthInfo.get(method))
        #print("DBUser.authenticate: authenticator:", a)
        if a is None or not a.enabled():
            return False, "Authenticated method disabled", None
        result, reason, expiration = a.authenticate(self, presented)
        #print(f"BaseDBUser.authenticate({method}):", result)
        return result, reason, expiration
        
    def set_password(self, realm, password, hashed=False):
        auth_info = self.AuthInfo.get("password", {})
        a = authenticator("password", realm, auth_info)
        self.AuthInfo["password"] = a.update_auth_info(self.Username, password, hashed=hashed)
        
    def get_password(self, realm):
        #print("DBUser.get_password(", realm, "): info:", self.AuthInfo)
        return self.AuthInfo.get("password", {}).get(realm)
        
    def get_dns(self):
        return self.AuthInfo.get("x509", [])

    def set_dns(self, dn_list):
        self.AuthInfo["x509"] = dn_list[:]

    def auth_method_enabled(self, method):
        return method in self.AuthInfo
        
    def ___auth_method_enabled(self, method, config):
        a = authenticator(method, config, self.AuthInfo.get(method))
        return a is not None and a.enabled()
        #return self.authenticator(method).enabled()

    @staticmethod
    def get(db, username):
        c = db.cursor()
        columns = BaseDBUser.columns("u")
        c.execute(f"""select {columns}, array(select ur.role_name from users_roles ur where ur.username=u.username)
                        from users u
                        where u.username=%s""",
                (username,))
        tup = c.fetchone()
        if not tup: return None
        (username, name, email, flags, auth_info, auid, roles) = tup
        u = BaseDBUser(db, username, name, email, flags, auth_info, auid)
        u.RoleNames = roles
        return u
        
    def is_admin(self):
        return "a" in (self.Flags or "")
    
    @staticmethod 
    def list(db):
        columns = BaseDBUser.columns("u")
        c = db.cursor()
        c.execute(f"""select {columns}, array(select ur.role_name from users_roles ur where ur.username=u.username)
            from users u
            order by u.username
        """)
        for username, name, email, flags, auth_info, auid, roles in c.fetchall():
            u = BaseDBUser(db, username, name, email, flags, auth_info, auid)
            u.RoleNames = roles
            #print("DBUser.list: yielding:", u)
            yield u

    @property
    def roles(self):
        return DBManyToMany(self.DB, "users_roles", "role_name", username = self.Username)
        
    def add_role(self, role):
        self.roles.add(role.Name if isinstance(role, BaseDBRole) else role)

    def remove_role(self, role):
        self.roles.remove(role.Name if isinstance(role, BaseDBRole) else role)


class BaseDBRole(DBObject):

    Table = "roles"
    Columns = ["name", "description"]
    PK = ["name"]

    def __init__(self, db, name, description=None, users=[]):
        DBObject.__init__(self, db)
        self.Name = name
        self.Description = description
            
    def __str__(self):
        return "[BaseDBRole %s %s]" % (self.Name, self.Description)
        
    __repr__ = __str__

    @property
    def members(self):
        return DBManyToMany(self.DB, "users_roles", "username", role_name=self.Name)

    @transactioned
    def save(self, transaction=None):
        transaction.execute("""
            insert into roles(name, description) values(%s, %s)
                on conflict(name) 
                    do update set description=%s
            """,
            (self.Name, self.Description, self.Description))
        return self
        
    @staticmethod
    def get(db, name):
        c = db.cursor()
        c.execute("""select r.description
                        from roles r
                        where r.name=%s
        """, (name,))
        tup = c.fetchone()
        if not tup: return None
        (desc,) = tup
        return BaseDBRole(db, name, desc)
        
    @staticmethod 
    def list(db, user=None):
        c = db.cursor()
        if isinstance(user, BaseDBUser):    user = user.Username
        if user:
            c.execute("""select r.name, r.description
                        from roles r
                            inner join users_roles ur on ur.role_name=r.name
                    where ur.username = %s
                    order by r.name
            """, (user,))
        else:
            c.execute("""select r.name, r.description
                            from roles r
                            order by r.name""")
        
        out = [BaseDBRole(db, name, description) for  name, description in fetch_generator(c)]
        #print("DBRole.list:", out)
        return out
        
    def add_member(self, user):
        self.members.add(user)
        return self
        
    def remove_member(self, user):
        self.members.remove(user)
        return self
        
    def set_members(self, users):
        self.members.set(users)
        return self
        
    def __contains__(self, user):
        if isinstance(user, BaseDBUser):
            user = user.Username
        return user in self.members
        
    def __iter__(self):
        return self.members.__iter__()



class Transaction(object):
    
    def __init__(self, connection, on_delete="commit"):
        assert not on_delete or on_delete in ("commit", "rollback")
        self.Connection = connection
        self.Cursor = self.Connection.cursor()
        self.InTransaction = False
        self.Exc = None
        self.Failed = False
        self.OnDelete = on_delete

    def begin(self):
        self.Cursor.execute("begin")
        self.InTransaction = True
        
    def commit(self):
        self.Cursor.execute("commit")
        self.InTransaction = False
        
    def rollback(self):
        self.Cursor.execute("rollback")
        self.InTransaction = False
        
    def execute(self, *params, **args):
        if not self.InTransaction:
            raise RuntimeError("Not in transaction")
        try:
            self.Cursor.execute(*params, **args)
        except:
            self.rollback()
            raise

    def executemany(self, *params, **args):
        if not self.InTransaction:
            raise RuntimeError("Not in transaction")
        try:
            self.Cursor.execute(*params, **args)
        except:
            self.rollback()
            raise

    def copy_from(self, *params, **args):
        if not self.InTransaction:
            raise RuntimeError("Not in transaction")
        try:
            self.Cursor.copy_from(*params, **args)
        except:
            self.rollback()
            raise

    def __enter__(self):
        self.begin()
        return self
        
    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is not None:
            self.Exc = (exc_type, exc_value, traceback)
            self.rollback()
        #else:  do not do commit here. Do it in the __del__ method instead
        #    self.commit()
            
    def __getattr__(self, name):
        return getattr(self.Cursor, name)
        
    def one(self):
        return self.Cursor.fetchone()

    def cursor_iterator(self):
        while True:
            tup = self.one()
            if tup is None:
                break
            yield tup

    results = cursor_iterator       # alias for clarity

    def __iter__(self):
        return self.results()
        
    def __del__(self):
        #print("Transaction __del__")
        if self.InTransaction:
            if self.OnDelete == "commit":
                self.commit()
            elif self.OnDelete == "rollback":
                self.rollback()

class ConnectionWithTransactions(object):
    
    def __init__(self, conn):
        self.Connection = conn
        
    def transaction(self, **args):
        return Transaction(self.Connection, **args)
        
    def __getattr__(self, name):
        return getattr(self.Connection, name)
